/* ====================================================================
  - Copyright (c) 1995-2000 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 * Author           : Takato Satsuma <takato@nowhere-land.org>
 * Created On       : 2002.02.05
 * Status           : Functional
 * Last Modified By : 2003.04.24
 *
 * NAME
 *     mod_vhost_limit.c
 *
 * PURPOSE 
 *     Limit (limit) access by one of the following method:
 *       * To restrict the number of simultaneous connections per
 *         a virtual host.
 *
 * HISTORY
 *     2002.02.05
 *       Initial revision 0.1
 *
 *     2002.11.19
 *       version 0.2. 
 *       * 'MaxConnPerVhost' directive was obsoleted.
 *         By using 'MaxVhostClients', it's able to control connections 
 *         per a virtual context.
 *
 *     2003.02.27
 *       Version 0.3.
 *       * Run as the post-read-request module. [Takato]
 *       * Returns an error if ExtendedStatus is not set to On. 
 *       * Takes into account state SERVER_BUSY_LOG and SERVER_BUSY_DNS.
 *       * Better accounts the current vhost connections. 
 *       * Changed comparison done on vhost names to a raw comparison of
 *         pointers on server_rec.
 *             above all by Mathias Herberts <Mathias.Herberts@iroise.net>
 *
 *     2003.04.24
 *       Version 0.4.
 *       * Add a flag 'APLOG_NOERRNO' not to print a previous error. [takato]
 *
 */

#define VERSION     "0.4"
#define VHOST_LIMIT_DEBUG	0

#define	CORE_PRIVATE
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_request.h"
#include "http_conf_globals.h"
#include "scoreboard.h"

typedef enum {
    VHOST_LIMIT_UNSET,
    VHOST_LIMIT_SET
} vhost_limit_mode;

typedef struct vhost_limit_conf {
    int max_clients;
    vhost_limit_mode mode;
} vhost_limit_conf;

module MODULE_VAR_EXPORT vhost_limit_module;

static void *
create_vhost_limit_config(pool *p, server_rec *s)
{
    vhost_limit_conf *conf;
    
    conf = (vhost_limit_conf *) ap_pcalloc(p, sizeof(vhost_limit_conf));
    conf->max_clients = HARD_SERVER_LIMIT;
    conf->mode = VHOST_LIMIT_UNSET;

    return conf;
}

static const char *
max_vhost_clients_set(cmd_parms *cmd, void *dummy, char *arg)
{
    int val;
    vhost_limit_conf *conf;
    
    /* We need extended status to have the vhost in the scoreboard */
    if (!ap_extended_status)
        return "ExtendedStatus must be set to On for mod_vhost_limit to be available";

    val = atoi(arg);
    conf = (vhost_limit_conf *) ap_get_module_config(cmd->server->module_config,
                &vhost_limit_module);

    if (val < 1)
        return "MaxVhostClients must be an integer greater than 0";
    else {
        conf->max_clients = val;
        conf->mode = VHOST_LIMIT_SET;
        ap_log_error(APLOG_MARK, APLOG_NOTICE|APLOG_NOERRNO, cmd->server,
            "mod_vhost_limit: %s MaxVhostClients = %d",
            cmd->server->server_hostname, val);
    }

    return NULL;
}

static const char *
obsoleted_warn(cmd_parms *cmd, void *dummy, char *arg)
{

    return "mod_vhost_limit> 'MaxConnPerVhost' is obsoleted. "
           "Use 'MaxVhostClients' instead.";
}

static const command_rec vhost_limit_cmds[] = {
    {"MaxVhostClients", max_vhost_clients_set, NULL, RSRC_CONF, TAKE1,
     "'MaxVhostClients' followed by non-negative integer"},
    {"MaxConnPerVhost", obsoleted_warn, NULL, RSRC_CONF, TAKE1,
     "'MaxConnPerVhost' followed by non-negative integer"},
    {NULL}
};

static int
check_vhost_limit(request_rec *r)
{
    int i, my_child_num, matched;
    short_score score_record;
    parent_score ps_record;
    server_rec *vhost;
    const char *req_host;
    vhost_limit_conf *conf;
#ifdef OPTIMIZE_TIMEOUTS
    vtime_t req_vtime;
#else
    time_t req_last_used;
#endif

    conf = (vhost_limit_conf *) ap_get_module_config(r->server->module_config,
                 &vhost_limit_module);

    /* if not configured for this vhost, just ignored */
    if (conf->mode == VHOST_LIMIT_UNSET)
        return OK;

    /* check if the scoreboard exists */
    if (!ap_exists_scoreboard_image()) {
        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, r,
            "mod_vhost_limit unavailable in inetd mode");
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    ap_sync_scoreboard_image();

    req_host = ap_get_server_name(r);

    my_child_num = r->connection->child_num;

#ifdef OPTIMIZE_TIMEOUTS
    req_vtime = ap_scoreboard_image->servers[my_child_num].cur_vtime;
#else
    req_last_used = ap_scoreboard_image->servers[my_child_num].last_used;
#endif

#if VHOST_LIMIT_DEBUG
    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_WARNING, r->server,
        "req_host:%s limited to %d connections", req_host, conf->max_clients);
#endif

    /* see the scoreboard */
    for (matched = 0, i = 0; i < HARD_SERVER_LIMIT; ++i) {
        if (my_child_num == i) {
            matched++;
            continue;
        } else {
            score_record = ap_scoreboard_image->servers[i];
            ps_record = ap_scoreboard_image->parent[i];
            vhost = score_record.vhostrec;
        }

        if (ps_record.generation != ap_my_generation)
            vhost = NULL;

        /* if vh is not NULL it's valid in this child. */
        if (!vhost)
            continue;

        if ((r->server == vhost) &&
                (score_record.status == SERVER_BUSY_READ ||
                 score_record.status == SERVER_BUSY_WRITE ||
                 score_record.status == SERVER_BUSY_LOG ||
                 score_record.status == SERVER_BUSY_DNS ||
                 score_record.status == SERVER_BUSY_KEEPALIVE)) {

            short slot_predates = 0;

#ifdef OPTIMIZE_TIMEOUTS
            if (req_vtime >= score_record.cur_vtime) {
                slot_predates = 1;
            }
#else
            if (req_last_used >= score_record.last_used) {
                slot_predates = 1;
            }
#endif

            /*
             * Evaluate the number of active connections for the vhost
             *
             * SERVER_BUSY_READ      No request processed yet, take into
             *                       account only if connection predates
             *                       that of request.
             * SERVER_BUSY_KEEPALIVE At least one request processed, take
             *                       into account systematically.
             * SERVER_BUSY_WRITE     Received request, about to answer.
             * SERVER_BUSY_LOG       Take into account systematically if
             * SERVER_BUSY_DNS       conn_count > 1, otherwise only if 
             *                       connection predates that of request.
             */

            switch (score_record.status) {
                case SERVER_BUSY_KEEPALIVE:
                    matched++;
                    break;
                case SERVER_BUSY_READ:
                    if (slot_predates) {
                        matched++;
                    }
                    break;
                 case SERVER_BUSY_WRITE:
                 case SERVER_BUSY_LOG:
                 case SERVER_BUSY_DNS:
                    if (score_record.conn_count > 1) {
                        matched++;
                    } else if (slot_predates) {
                        matched++;
                    }
                    break;
                 default:
            }
        }
    }

#if VHOST_LIMIT_DEBUG
    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_WARNING, r->server,
        "req_host matched %d connections", matched);
#endif

    if (matched > conf->max_clients) {
        ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
            "client access to %s deferred, MaxVhostClients %d exceeded",
            req_host, conf->max_clients);
        return HTTP_SERVICE_UNAVAILABLE;
    }

    return OK;
}

module MODULE_VAR_EXPORT vhost_limit_module =
{
    STANDARD_MODULE_STUFF,
    NULL,                               /* initializer */
    NULL,                               /* dir config creater */
    NULL,                               /* dir merger: default is to override */
    create_vhost_limit_config,          /* server config */
    NULL,                               /* merge server config */
    vhost_limit_cmds,                   /* command table */
    NULL,                               /* handlers */
    NULL,                               /* filename translation */
    NULL,                               /* check_user_id */
    NULL,                               /* check auth */
    NULL,                               /* check access */
    NULL,                               /* type_checker */
    NULL,                               /* fixups */
    NULL,                               /* logger */
    NULL,                               /* header parser */
    NULL,                               /* child_init */
    NULL,                               /* child_exit */
    check_vhost_limit                   /* post read-request */
};

/* 
$Id: mod_vhost_limit.c,v 1.10 2003/04/23 16:50:36 takato Exp $
 */
